import asyncio
import os
import shutil
from pathlib import Path
import subprocess
from py7zr import unpack_7zarchive
import orjson
import pickle
import numpy as np
import pandas as pd
import aiofiles
import io
from dl2050utils.core import listify, xre

Path.ls = lambda x: list(x.iterdir())

# ################################################################################
# pickle_save (pickle_dump), pickle_load
# np_save, np_load, np_load_async
# df_save, df_load
# json_load (json_parse), json_dump, read_json
# json_saves, json_loads
# ################################################################################

def pickle_save(p, d):
    p = Path(p)
    p = p.with_suffix('.pickle')
    try:
        with open(p, 'wb') as f:
            pickle.dump(d, f)
        return 0
    except Exception:
        return 1

# Deprecated
def pickle_dump(p, d): return pickle_save(p, d)
    
def pickle_load(p):
    p = Path(p)
    p = p.with_suffix('.pickle')
    if not p.is_file():
        return None
    try:
        with open(p, 'rb') as f:
            d = pickle.load(f)
        return d
    except Exception:
        return None
    
def np_save(p, d, allow_pickle=True):
    p = Path(p)
    p = p.with_suffix('.npy')
    try:
        np.save(p, d, allow_pickle=allow_pickle)
        return 0
    except Exception:
        return 1
    
def np_load(p, allow_pickle=True):
    p = Path(p)
    p = p.with_suffix('.npy')
    if not p.is_file():
        return None
    try:
        if allow_pickle:
            d = np.load(p, allow_pickle=allow_pickle)
            if d is None: return None
            return d.item()
        else:
            return np.load(p)
    except Exception:
        return None
    
async def np_load_async(p, allow_pickle=False):
    p = Path(p)
    p = p.with_suffix('.npy')
    if not p.is_file():
        return None
    try:
        async with aiofiles.open(p, mode='rb') as f:
            buffer = await f.read()
        f = io.BytesIO(buffer)
        if allow_pickle:
            d = np.load(f, allow_pickle=allow_pickle)
            if d is None: return None
            return d.item()
        else:
            return np.load(f, allow_pickle=allow_pickle)
    except Exception:
        return None
    
def df_save(p, df):
    p = Path(p)
    p = p.with_suffix('.feather')
    try:
        df.to_feather(p)
        return 0
    except Exception:
        return 1
    
def df_load(p):
    p = Path(p)
    p = p.with_suffix('.feather')
    if not p.is_file():
        return None
    try:
        df = pd.read_feather(p)
        return df
    except Exception:
        return None

def json_save(p, d):
    p = Path(p)
    p = p.with_suffix('.json')
    try:
        with open(p, 'wb') as f:
            f.write(orjson.dumps(d))
        return 0
    except Exception:
        return 1
    
def json_load(p):
    p = Path(p)
    p = p.with_suffix('.json')
    if not p.is_file():
        return None
    try:
        with open(p, 'rb') as f:
            return orjson.loads(f.read())
    except Exception:
        return None

def json_dumps(o):
    try:
        return orjson.dumps(o, option=orjson.OPT_SERIALIZE_NUMPY).decode()
    except:
        return None

def json_loads(s):
    if s is None or s=='': return None
    try:
        o = orjson.loads(s)
        if o=={}: return None
        return o
    except:
        return None

# To be reviwed (used in PW) and renamed to json_load
def read_json(p): return json_load(p)

# ################################################################################
# Shell commands:
#   cp, rm, sh_run, run_asyc
# ################################################################################

def cp(p1, p2):
    """ Copies recursively p1 to p2 with overwrite. If p1 is a directory, copies recursively. """
    p1,p2 = Path(p1),Path(p2)
    if not p1.exists():
        print(f'File {p1} not found')
        return 1
    try:
        if p1.is_file():
            shutil.copyfile(p1, p2)
            return 0
        if p2.is_dir():
            shutil.rmtree(p2)
        shutil.copytree(p1, p2)
    except Exception:
        return 1
    
def rm(p):
    " Removes file or directory p. If p is a directory, removes recursively. "
    try: 
        p = Path(p)
        if p.is_file():
            p.unlink()
            return 0
        if p.is_dir():
            return shutil.rmtree(p)
    except Exception:
        return 1

def sh_run(cmd):
    """ Executas a shell command and return a tuple with exit code and stdout. """
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (stdout, stderr) = proc.communicate()
    stdout, stderr = stdout.decode("utf-8"), stderr.decode("utf-8")
    if proc.returncode!=0: raise RuntimeError(f'sh_run command \n{cmd}\n exit with error code {proc.returncode}:\n{stderr}')
    return (proc.returncode, stdout)

async def run_asyc(*args):
    try:
        process = await asyncio.create_subprocess_exec(*args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await process.communicate()
        return process.returncode, stdout.decode(), stderr.decode()
    except:
        return -1, '', 'Command not found'
    
# ################################################################################
# get_all_files, get_dir_files, get_dir_dirs
# ################################################################################

def get_all_files(p, filetypes=None):
    p = Path(p)
    filetypes = [e.lower() for e in listify(filetypes)]
    filetypes = [f'.{e}' if e[0]!='.' else e for e in filetypes]
    files = [p1 for p1 in p.rglob('*')]
    files = [e for e in files if xre('(DS_Store)',str(e)) is None and xre('(__MACOSX)',str(e)) is None]
    if len(filetypes): files = [f for f in files if f.suffix.lower() in filetypes]
    return files

# Draft
def get_dir_files(path, types=None):
    return [f for f in path.glob('**/*') if f.is_file() and (types==None or f.suffix[1:] in types)]

# Draft
def get_dir_dirs(path):
    return [d for d in path.glob('**/*') if d.is_dir()]
    
# ###################################################################################
# zip_get_registered_formats, zip_register_formats, iszip, dozip, unzip, unzip_tree
# ###################################################################################

def zip_get_registered_formats():
    return sum([e[1] for e in shutil.get_unpack_formats()], [])
    
def zip_register_formats():
    zips = zip_get_registered_formats()
    if '.7z' not in zips:
        print('Registering .7z')
        shutil.register_unpack_format('7zip', ['.7z'], unpack_7zarchive)
    if '.hsz' not in zips:
        print('Registering .hsz')
        shutil.register_unpack_format('7zip', ['.hsz'], unpack_7zarchive)

def dozip(p, remove=False):
    p = Path(p)
    if not p.exists():
        print(f'dozip: file not found: {p}')
        return 1
    try: 
        shutil.make_archive(p, 'zip', p)
        if remove:
            if rm(p): return 1
        return 0
    except Exception:
        return 1
    
def iszip(p):
    zipext = zip_get_registered_formats()
    archext = ['.rar']
    return (Path(p)).suffix in zipext+archext
    
def unzip(p, remove=False):
    """ Unzip or untar a file. If it is not an archive does nothing. """
    p = Path(p)
    if not p.exists():
        print(f'unzip: file not found: {p}')
        return 1
    if p.is_dir():
        return 0
    if not iszip(p):
        return 0
    if p.suffix in ['.rar']:
        code,stdout = sh_run(f'unrar x {p} {p.parent}')
        if code!=0:
            print(f'unrar error: {stdout}')
            return 1
        if remove:
            if rm(p): return 1
        print(f'File {p} extracted from archive')
        return 0
    p_dir = p.with_suffix('')
    try: 
        shutil.unpack_archive(p, p_dir)
        if remove:
            if rm(p): return 1
        print(f'File {p} unzipped')
        return 0
    except Exception as exc:
        print(f'Exception: {exc}')
        return 1
    
def unzip_tree(p):
    """ Get all subfolders and files from and unzip/untar if needed. """
    for p1 in [p1 for p1 in get_all_files(p)]:
        if unzip(p1, remove=True):
            print(f'unzip_tree: error unzipping {p1}')



"""
restapp.py
App Class

Defines the logic for the REST App, including startup, shudown, Auth routes, downstream services routes and Custom Application routes.

The Auth routes are defined in the Auth module.

The downstream services routes include:
    /api/contact
    /api/get_meta
    /api/download_req
    /api/upload_req
    /api/download_workfile/{fname}
    /api/upload_workfile
    /api/publish_job
    /api/get_jobs
    /api/get_job

The Custom Application are defined in the routes dict with the following structure:
    routes = {
        <route_path_string>: <route_defenition>,
        ...
    }
    <route_defenition> is a dict with keys 'f','args' and 'kwargs', with the function name implementing the route along with
    corresponding args and kwargs as lists of strings.

    For example:
        routes = {
            '/api/health': {'f': health, 'args':[], 'kwargs':[]},
            ...
        }

Every custom function implementing the route must observe the following:
    1) All inputs parsed from the Request as function parametes that must be defined in the lists args and kwargs above;
    2) All returns should be a dict with the format {status:<status>, ...}:
        - <status> should be either 'OK' or 'ERROR'
        - If 'status':'OK' extra attributes may be present and will be sent to the client
        - If 'status':'ERROR', attribute 'error_msg' should be present, optionally along with 'error_code'
        - If status is not present, 'status':'OK' is automatically inserted;
        - If the return is not a dict, a dict is created with {'status':'OK', 'result':<return>}
        - If None is returned, a dict is created with only {'status':'OK'}
        Examples:
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
            return {'data':[arg1, kwarg1]} # Client gets {'status':'OK', 'data':[arg1, kwarg1]}
            return [1,2,3] # Client gets {'status':'OK', 'result':[1, 2, 3]}
            return None # Client gets {'status':'OK'}
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
    3) All normal returns are converted into HTTP responses with HTTP status 200 and the return data as JSON
    4) It is possible to raise an exception through log_and_raise_app(msg), that is converted in an error return
       of the form {'status':'ERROR', 'error_type':'APP', 'error_msg':<msg>} and then sent as an HTTP response
       with HTTP status 400 and the error data as JSON
    5) If the function has an uncatched exception (bug), the App catches it and converts in the following response:
       {'status':'ERROR', 'error_type':'EXCEPTION', 'error_msg': <exception_details>} and then sent as an HTTP response
       with HTTP status 400 and the error data as JSON

Regarding files upload/downloand, consider the following:
    1) The routes /api/download_req and /api/upload_req provice signed URLs for downnload/upload.
       The signed URLs can be obtained from the File Server or Google Cloud Storage.
       They are intended to offload the REST server of dealing with files, proxying to external services.
       The user namespace is limited to its email prefix: <bucket>/<email>.
    2) The routes /api/download_workfile and /api/upload_workfile provides downnload/upload functionality for
       registered users.
       User have isolated namespace under '$DATAPATH/workfiles/<email>'

The remaing methods deal with jobs lifecycle and there is a method to register contact requests.
"""

import time
from datetime import datetime, timezone
from pathlib import Path
from starlette.responses import FileResponse
from starlette.routing import Route
from starlette.authentication import requires
from dl2050utils.core import oget
from dl2050utils.restutils import HTTPException, log_and_raise_exception, log_and_raise_rest, log_and_raise_service
from dl2050utils.restutils import rest_ok,enforce_required_args, get_optional_args, get_meta, form_files_upload
from dl2050utils.restutils import fs_download_signed_url, fs_upload_signed_url, gcs_download_signed_url, gcs_upload_signed_url

class App():

    def __init__(self, cfg, LOG, NOTIFY, db, mq, auth, path, routes=[], appstartup=None, perm=None):
        self.service = cfg['service']
        self.path = path
        self.cfg,self.LOG,self.NOTIFY,self.db,self.mq,self.auth = cfg,LOG,NOTIFY,db,mq,auth
        self.routes,self.appstartup,self.perm = routes,appstartup,perm
        self.d = {'LOG':LOG, 'path':path, 'db':db, 'mq':mq}
        self.fs_secret = oget(self.cfg,['fs','secret'])
        if self.fs_secret is None:
            self.LOG(3, 0, label='APP', label2='REST',  msg='fs_secret not found')

    # ####################################################################################################
    # Helper methods
    # ####################################################################################################

    async def db_increase(self, tbl, k, kv, c, prefix=''):
        row = await self.db.select_one(tbl, {k: kv})
        if row is None or len(row)==0:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: select')
        row[c] += row[c]+1
        res = await self.db.update(tbl, row)
        if res:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: update')

    # ####################################################################################################
    # Internal methods: startup, shutdown, get_routes
    # ####################################################################################################

    async def startup(self):
        model,meta = oget(self.cfg,['app','model']),None
        if model is not None: meta = await get_meta(self.path, self.db, model)
        self.d['meta'] = meta
        if self.appstartup is None: return False
        return await self.appstartup(self.d)

    def shutdown(self):
        self.LOG(2, 0, label='APP', label2='shutdown', msg='OK')
        return False   

    def get_routes(self):
        BASE_ROUTES = [
            Route('/api/contact', endpoint=self.contact, methods=['POST']),
            Route('/api/get_meta', endpoint=self.get_meta, methods=['GET']),
            Route('/api/download_req', endpoint=self.download_req, methods=['POST']),
            Route('/api/upload_req', endpoint=self.upload_req, methods=['POST']),
            Route('/api/download_workfile/{fname:path}', endpoint=self.download_workfile, methods=['GET']),
            Route('/api/upload_workfile/{fname:path}', endpoint=self.upload_workfile, methods=['POST']),
            Route('/api/publish_job', endpoint=self.publish_job, methods=['POST']),
            Route('/api/get_jobs', endpoint=self.get_jobs, methods=['POST']),
            Route('/api/get_job', endpoint=self.get_job, methods=['POST']),
        ]
        APP_ROUTES = [Route(e, endpoint=self.app_route, methods=['POST']) for e in self.routes]
        return BASE_ROUTES + APP_ROUTES
    
    # ####################################################################################################
    # REST methods for Apps
    # ####################################################################################################

    async def contact(self, request):
        """
        """
        data = await request.json()
        d = enforce_required_args(self.LOG, data, ['email','name','msg'], label='AUTH', label2='check_user')
        d['ts']: datetime.now(timezone.utc)
        err = await self.db.insert('contacts', d)
        if err:
            log_and_raise_service(self.LOG, label='REST', label2='contact', msg='DB access error')
        html = '<h1>Name</h1><p>{d["email"]</p><h1>Email</h1><p>{d["email"]</p><h1>Message</h1><p>{d["msg"]</p>}'
        self.notify.send_mail_async(d['email'], subject='Contact from Website (cardiolife.global)', html=html)
        return rest_ok()

    @requires('authenticated')
    async def get_meta(self, request):
        """
            Return the meta information.
        """
        u = await self.auth.check_auth(request)
        return rest_ok(self.d['meta'])
    
    @requires('authenticated')
    async def upload_req(self, request):
        """
            Returns an upload signed URL for <bucket>/<email>/<key> for the specified key.
            The user namespace is limited to <bucket>/<email> and <email> is prepended to <key> as dir.
            The URL will be executed either by the File Server or the Cloud File Server (TO BE DEFINE HOW)
            A 24h default timeout is set.
            TODO: Return also the download request
            TODO: Client hostname is redirected to 10.0.1.3 -> Fix
            TODO: Include method to select the use of Local FS, FS, GCS. --> cfg
            TODO: Control size of user uploads
            TODO: Resumable uploads with GCE
        """
        if not self.fs_secret:
            log_and_raise_service(self.LOG, label='REST', label2='upload_req', msg='Upload secret not defined')
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        # self.db_increase('users', 'uid', u['uid'], 'uploads', 'upload_req')
        data = await request.json()
        FILE_SERVER_PROVIDER = 'FS'
        args = ['bucket','key','size']
        args = enforce_required_args(self.LOG, data, args, label='REST', label2='upload_req')
        kwargs = get_optional_args(data, ['fs_url'])
        self.perm(self.d, u, request.url.path, args)
        key = f"{email}/{args['key']}"
        timeout = 7*24*3600
        url = None
        FILE_SERVER_PROVIDER = 'FS'
        if FILE_SERVER_PROVIDER=='FS':
            # request.headers['Host']
            # https://10.0.1.3/
            # FILE_SERVER_BASE_URL = 'https://cardiolife.global' # request.url.host, request.client.host
            fs_url = kwargs['fs_url'] or 'cardiolife.global'
            FILE_SERVER_BASE_URL = f'https://{fs_url}'
            url = fs_upload_signed_url(self.fs_secret, FILE_SERVER_BASE_URL, args['bucket'], key, args['size'], timeout=timeout)
        else:
            url = gcs_upload_signed_url(args['bucket'], args['key'], timeout=timeout, size=args['size'])
        self.LOG(2, 0, label='APP', label2='upload_req', msg={'url':url})
        # return rest_ok({'url':url, 'info':{'request_client':request.client, 'request_url':request.url}})
        return rest_ok({'url':url, 'info':{'Host':oget(request.headers,['Host'])}})
    
    @requires('authenticated')
    async def download_req(self, request):
        """
            Returns a download signed URL for <bucket>/<email>/<key> for the specified key.
            The user namespace is limited to <bucket>/<email> and <email> is prepended to <key> as dir.
            The URL will be executed either by the File Server or the Cloud File Server (TO BE DEFINE HOW)
            A 24h default timeout is set.
            TODO: Client hostname is redirected to 10.0.1.3 -> Fix
            TODO: Include method to select the use of Local FS, FS, GCS. --> cfg
        """
        if not self.fs_secret:
            log_and_raise_service(self.LOG, label='REST', label2='download_req', msg='Upload secret not defined')
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        # self.db_increase('users', 'uid', u['uid'], 'downloads', 'download_req')
        data = await request.json()
        args = ['bucket','key']
        args = enforce_required_args(self.LOG, data, args, label='REST', label2='download_req')
        kwargs = get_optional_args(data, ['fs_url'])
        self.perm(self.d, u, request.url.path, args)
        bucket = args['bucket']
        key = f"{email}/{args['key']}"
        timeout = 7*24*3600
        url = None
        FILE_SERVER_PROVIDER = 'FS'
        if FILE_SERVER_PROVIDER=='FS':
            # https://10.0.1.3/
            # FILE_SERVER_BASE_URL = 'https://cardiolife.global' # request.url.host, request.client.host
            fs_url = kwargs['fs_url'] or 'cardiolife.global'
            FILE_SERVER_BASE_URL = f'https://{fs_url}'
            url = fs_download_signed_url(self.fs_secret, FILE_SERVER_BASE_URL, bucket, key, timeout=timeout)
        else:
            url = gcs_download_signed_url(bucket, key, timeout=timeout)
        self.LOG(2, 0, label='APP', label2='download_req', msg={'url':url})
        return rest_ok({'url':url})

    @requires('authenticated')
    async def download_workfile(self, request):
        """
            Downloads a workfile file fname belonging to a registered user.
            Workfiles are usually results produced by workers (throught MQ).
            Workfiles are stored in the folder '$DATAPATH/workfiles/<service>/<email>'.
            Only the user can access the file.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        fname = request.path_params['fname']
        p = Path(f'/data/{self.service}/workfiles/{email}/{fname}')
        if not p.is_file():
            log_and_raise_service(self.LOG, label='REST', label2='download_workfile', msg=f'File not found: {str(p)}')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=str(p))

    # TODO : NOT FINISHED
    @requires('authenticated')
    async def upload_workfile(self, request):
        """
            Uploads a workfile file fname belonging to a registered user.
            Workfiles for upload are usually content produced by the users.
            Workfiles are stored in the folder '$DATAPATH/workfiles/<service>/<email>'.
            Only the user can access the file.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        fname = request.path_params['fname']
        p = Path(f'/data/{self.service}/workfiles/{email}/{fname}')
        max_size = int(1e9)
        form = await request.form()
        if form is None:
            log_and_raise_service(self.LOG, label='REST', label2='upload_workfile', msg=f'Upload form error')
        await form_files_upload(self.LOG, form.items(), p, max_size=max_size)
        return rest_ok()

    @requires('authenticated')
    async def publish_job(self, request):
        """
            Publishes a new job, calling mq.publish after checking permissions.
            Jobs require the user email as an attribute.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['q','payload'], label='REST', label2='publish_job')
        q,payload = args['q'],args['payload']
        self.perm(self.d, u, request.url.path, args)
        jid = await self.mq.publish(q, email, payload)
        if jid is None :
            log_and_raise_service(self.LOG, label='REST', label2='publish_job', msg='MQ publish error')
        self.LOG(2, 0, label='APP', label2='publish_job')
        return rest_ok({'jid': jid})

    @requires('authenticated')
    async def get_jobs(self, request):
        """
            Gets all jobs from a user, optionaly filtered by qname, pending or jobs not done.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        kwargs = get_optional_args(data, ['qname','pending','not_done'])
        jobs = await self.mq.get_jobs(email, **kwargs)
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_job(self, request):
        """
            Gets a job details, identified by jid.
        """
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['jid'], label='REST', label2='get_job')
        job = await self.mq.get_job(args['jid'])
        return rest_ok(job)

    @requires('authenticated')
    async def app_route(self, request):
        """
            Implements all the custom app routes.
        """
        if request.url.path not in self.routes:
            log_and_raise_rest(self.LOG, label='REST', label2=f'app_route {request.url.path}', msg='url path not found')
        d = self.routes[request.url.path]
        cb,args,kwargs = d['f'],d['args'],d['kwargs']
        u = await self.auth.check_auth(request)
        data = await request.json()
        args2 = enforce_required_args(self.LOG, data, args, label='REST', label2='app_route')
        kwargs2 = get_optional_args(data, kwargs)
        self.perm(self.d, u, request.url.path, {**args2, **kwargs2})
        try:
            res = await cb(self.d, u, *[args2[e] for e in args2], **kwargs2)
        except HTTPException as exc:
            raise
        except Exception as exc:
            log_and_raise_exception(self.LOG, label='REST', label2='app_route', msg=str(exc))
        return rest_ok(res)



import datetime
import random
import urllib
from pathlib import Path
import json
import hashlib
import orjson
import jwt
import asyncpg
import aiohttp
import starlette
from starlette.responses import JSONResponse
from google.cloud import storage
from dl2050utils.core import oget
from dl2050utils.fs import read_json

# ####################################################################################################
# Utility functions
# ####################################################################################################

def mk_key(n=4):
    return ''.join([chr(48+i) if i<10 else chr(65-10+i) for i in [random.randint(0, 26+10-1) for _ in range(n)]])
    # return ''.join(random.choice(string.ascii_lowercase) for i in range(n))

def get_hash(o, secret):
    o1 = {**o}
    o1['secret']=secret
    return hashlib.sha224(json.dumps(o1).encode()).hexdigest()

def check_hash(o, h, secret):
    o1 = {**o}
    o1['secret']=secret
    return get_hash(o1,secret)==h

def mk_jwt_token(uid, email, secret):
    JWT_EXP_DELTA_SECONDS = 30*24*3600
    payload = {
        'uid':uid,
        'email':email,
        'username':'',
        'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=JWT_EXP_DELTA_SECONDS)
    }
    return jwt.encode(payload, secret, 'HS256') # .decode('utf-8')

# ####################################################################################################
# REST Responses and Exceptions
# ####################################################################################################
    
def orjson_serialize(obj):
    if isinstance(obj, asyncpg.pgproto.pgproto.UUID):
        return str(obj)
    raise TypeError

class OrjsonResponse(JSONResponse):
    def render(self, content):
        return orjson.dumps(content, option=orjson.OPT_SERIALIZE_NUMPY, default=orjson_serialize)

def rest_ok(d=None):
    if d is None: d = {}
    if type(d)!=dict: d = {'result':d}
    if 'status' not in d: d['status'] = 'OK'
    return OrjsonResponse(d)

class HTTPException(Exception):
    def __init__(self, status_code, detail, error_type='APP'):
        self.status_code,self.detail,self.error_type = status_code,detail or '',error_type
    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f'{class_name}(status_code={self.status_code!r}, detail={self.detail!r})'

def log_and_raise(LOG, status_code=400, error_type='APP', label='', label2='', msg=''):
    if LOG is not None:
        LOG(4, 0, label=label, label2=label2, msg=msg)
    raise HTTPException(status_code, msg, error_type=error_type)

def log_and_raise_exception(LOG, label='', label2='', msg=''):
    return log_and_raise(LOG, error_type='EXCEPTION', label=label, label2=label2, msg=msg)

def log_and_raise_rest(LOG, label='', label2='', msg=''):
    return log_and_raise(LOG, error_type='REST', label=label, label2=label2, msg=msg)

def log_and_raise_service(LOG, label='', label2='', msg=''):
    return log_and_raise(LOG, error_type='SERVICE', label=label, label2=label2, msg=msg)

def log_and_raise_app(msg=''):
    return log_and_raise(None, error_type='APP', msg=msg)

def rest_exception(LOG, label, label2, msg):
    LOG(4, 0, label=label, label2=label2, msg=msg)
    raise HTTPException(400, detail=f'EXCEPTION: {msg}')

# ####################################################################################################
# Args
# ####################################################################################################

def enforce_required_args(LOG, payload, args, label='', label2='', as_list=False):
    """ Returns dict (or list is as_list is True) with args present in payload. Raises REST exception if any arg is missing """
    args2,miss = {},[]
    if payload is None:
        return args2
    for e in args:
        if e not in payload:
            miss.append(e)
        else:
            args2[e] = payload[e]
    if len(miss):
        msg = f'Missing required args: {", ".join(miss)}'
        log_and_raise_rest(LOG, label=label, label2=label2, msg=msg)
    if as_list:
        return [args2[e] for e in args2]
    return args2

def get_optional_args(payload, args, as_list=False):
    """ Returns dict (or list is as_list is True) with args present in payload """
    if as_list:
        return [payload[e] for e in args if e in payload] if payload is not None else []
    return {e:payload[e] for e in args if e in payload} if payload is not None else {}

# ####################################################################################################
# Meta
# ####################################################################################################

async def get_meta(path, db, model):
    meta = read_json(f'{path}/{model}/{model}.json')
    if meta is not None:
        return meta
    row = await db.select_one('models', {'model': model})
    if row is not None:
        return json.loads(row['meta'])
    return None

def sync_get_meta(path, db, model):
    meta = read_json(f'{path}/{model}/{model}.json')
    if meta is not None:
        return meta
    row = db.sync_select_one('models', {'model': model})
    if row is not None: return json.loads(row['meta'])
    return None

# ####################################################################################################
# Testing REST service requests from from client:
#   rest_request, do_login
# ####################################################################################################

async def rest_request(session, url, headers, host, port, payload=None, method='POST', json=True):
    print(f'{method} url: http://{host}:{port}{url}')
    if method not in ['POST', 'GET']:
        print('Invalid HTTP method')
        return
    if method=='POST':
        payload = payload or {}
        r = await session.post(f'http://{host}:{port}{url}', headers=headers, json=payload)
    else:
        r = await session.get(f'http://{host}:{port}{url}', headers=headers)
    print(f'HTTP response status: {r.status}')
    if r.status==500:
        trace = await r.text()
        print('Server EXCEPTION trace: :', trace)
        return None
    else:
        if r.status==403:
            print(r)
            return None
        if r.status>=400:
            text = await r.text()
            print(text)
        if json:
            res = await r.json()
            return res
        else:
            return None
        
async def do_login(session, host, port, url, email, passwd):
    headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
    payload = {'email': email, 'passwd': passwd}
    res = await rest_request(session, url, headers, host, port, payload=payload, method='POST', json=True)
    jwt_token = oget(res,['jwt_token'])
    if jwt_token is None:
        print('ERROR: ', res)
        return None
    return {'Content-Type': 'application/json', 'Accept': 'application/json', 'Authorization':f"Bearer {jwt_token}"}

# ##############################################################################################################
# File download from url:
#   async_upload, async_download
# ##############################################################################################################

async def async_upload(url, ps):
    """Uploads files ps to url."""
    async with aiohttp.ClientSession() as session:
        data = aiohttp.FormData()
        for i,p in enumerate(ps):
            data.add_field(f'file-{i}', open(p,'rb'), filename=p)
        async with session.post(url, data=data) as r:
            if(r.status!=200):
                print(f'Error: status=={r.status}, msg={await r.text()}')
                return True
            print(await r.text())
            return False
        
async def async_download(url):
    """Downloads file from url and returns the file data."""
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as r:
            if(r.status!=200):
                print(f'Error: status=={r.status}, msg={await r.text()}')
                return None
            return await r.text()

# Draft
# def download_url(url, dest, overwrite=False, pbar=None, show_progress=True, chunk_size=1024*1024, timeout=4, retries=5):
#     """Downloads file from url to dest unless it exists and not overwrite."""
#     if os.path.exists(dest) and not overwrite: return
#     s = requests.Session()
#     s.mount('http://',requests.adapters.HTTPAdapter(max_retries=retries))
#     u = s.get(url, stream=True, timeout=timeout)
#     try:
#         file_size = int(u.headers["Content-Length"])
#     except:
#         show_progress = False
#     with open(dest, 'wb') as f:
#         nbytes = 0
#         # if show_progress: pbar = progress_bar(range(file_size), auto_update=False, leave=False, parent=pbar)
#         try:
#             for chunk in u.iter_content(chunk_size=chunk_size):
#                 nbytes += len(chunk)
#                 if show_progress: pbar.update(nbytes)
#                 f.write(chunk)
#         except requests.exceptions.ConnectionError as exc:
#             print(f'EXCEPTION: Download of {url} has failed after {retries} retries: {str(exc)}')
#             return False
#     return True

# Draft
# def download_file(url, fname=None, ext='.tgz'):
#     """Downloads file from url to destination fname."""
#     p = Path(fname or f"{url.split('/')[-1]}{ext}")
#     (p.parent).mkdir(parents=True, exist_ok=True )
#     if not p.exists():
#         print(f'Downloading {url}')
#         if not download_url(f'{url}{ext}', p):
#             return None
#     return p

# ################################################################################
# form_files_upload
# ################################################################################

async def form_files_upload(LOG, items, p, max_size=int(1e9)):
    n,sz = 0,0
    for k,f in items:
        if type(f)==starlette.datastructures.UploadFile:
            sz += f.size
            if sz>max_size:
                log_and_raise_service(LOG, label='FS', label2='/upload', msg=f'Files size exceded {max_size} limit')
            fname = urllib.parse.unquote(f.filename)
            p1 = Path(fname)
            # if fname[0]=='/': fname = fname[1:]
            p1 = p / f'{p1.stem}{p1.suffix}'
            p1.parent.mkdir(parents=True, exist_ok=True)
            data = await f.read()
            with open(str(p1), 'wb') as f2:
                f2.write(data)
            n += 1
    return n,sz

# ################################################################################
# Signed URLs for File Server download/upload: 
#   gcs_download_signed_url, gcs_upload_signed_url
# ################################################################################

def fs_download_signed_url(secret, base_url, bucket, key, timeout=7*24*3600):
    """
        baseurl must specify the protocol, example: http://fs:4000
    """
    payload = {
        'bucket': bucket,
        'key': key,
        'ts': datetime.datetime.now().isoformat(),
        'timeout': timeout
    }
    payload['h'] = get_hash(payload, secret)
    url = f'{base_url}/download?{urllib.parse.urlencode(payload)}'
    return url

def fs_upload_signed_url(secret, base_url, bucket, key, size, timeout=7*24*3600):
    """
        baseurl must specify the protocol, example: http://fs:4000
    """
    payload = {
        'bucket': bucket,
        'key': key,
        'size': size,
        'ts': datetime.datetime.now().isoformat(),
        'timeout': timeout
    }
    payload['h'] = get_hash(payload, secret)
    url = f'{base_url}/upload?{urllib.parse.urlencode(payload)}'
    return url

# ################################################################################
# Signed URLs for Google Cloud Storage download/upload: 
#   gcs_download_signed_url, gcs_upload_signed_url
# ################################################################################

def gcs_download_signed_url(bucket_name, blob_name, timeout=15*60):
    """
        Generates a v4 signed URL for downloading a blob.
        Signed url expires according to timeout (in seconds).
        Requires a service account key json file (obtained in IAM).
        Requires the ENV variable GOOGLE_APPLICATION_CREDENTIALS with this file path:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = path_to_file
        https://cloud.google.com/storage/docs/samples/storage-generate-signed-url-v4#storage_generate_signed_url_v4-python
    """
    try:
        sc = storage.Client()
        bucket = sc.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        url = blob.generate_signed_url(version="v4", expiration=datetime.timedelta(seconds=timeout), method="GET")
        return url
    except Exception as exc:
        print(f'EXCEPTION: {str(exc)}')
        return None

def gcs_upload_signed_url(bucket_name, blob_name, timeout=15*60, size=None):
    """
        Generates a v4 signed URL for uploading a blob using HTTP PUT.
        Signed url expires according to timeout (in seconds).
        Requires a service account key json file (obtained in IAM).
        Requires the ENV variable GOOGLE_APPLICATION_CREDENTIALS with this file path:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = path_to_file
        https://cloud.google.com/storage/docs/samples/storage-generate-upload-signed-url-v4
    """
    try:
        sc = storage.Client()
        bucket = sc.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        url = blob.generate_signed_url(version="v4", expiration=datetime.timedelta(seconds=timeout), method="PUT", content_type="application/octet-stream")
    except Exception as exc:
        print(f'EXCEPTION: {str(exc)}')
        return None
    return url

# ####################################################################################################
# Etc - to review
# ####################################################################################################

# def mk_weeks(ds1='2018-01-01', ds2=None, weekday=6):
#     d1 = datetime.datetime.strptime(ds1, '%Y-%m-%d').date()
#     delta = 5 - d1.weekday()
#     if delta<0: delta+=7
#     d1 += datetime.timedelta(days=delta)
#     d2 = datetime.datetime.now().date() if ds2 is None else datetime.datetime.strptime(ds2, '%Y-%m-%d').date()
#     ds = [d.strftime("%Y-%m-%d") for d in rrule.rrule(rrule.WEEKLY, dtstart=d1, until=d2)]
#     return ds[::-1]

# def get_week2(weeks, week): return weeks[weeks.index(week)+1] if weeks.index(week)+1<len(weeks) else None

# def s3_urls(s3, bucket_name, prefix):
#     response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix, MaxKeys=1024)['Contents']
#     return [f'http://{bucket_name}.s3-eu-west-1.amazonaws.com/{e["Key"]}' for e in response]



import re
import pandas as pd
import numpy as np
import matplotlib, matplotlib.pyplot as plt
import seaborn as sns
from dl2050utils.core import oget

# ####################################################################################################
# Utils
# ####################################################################################################

def df_convert_32(df):
    for c in df.columns:
        if df.dtypes[c]==np.int64: df[c] = df[c].astype(np.int32)
        if df.dtypes[c]==np.float64: df[c] = df[c].astype(np.float32)

def df_mem(df):
    m = pd.DataFrame([(e[0],df[e[0]].dtypes,e[1]/2**30) for e in df.memory_usage(deep=True)[1:].items()], columns=['col','type','size'])
    display(m)
    print(f"Total {m['size'].sum():.1f}Gb, #{len(df)/1e6:.1f}M")

# ####################################################################################################
# df_join_by_range_2
# ####################################################################################################

def df_join_by_range_2(df, df_range, df_range_key, key_c, v_c, v1_c, v2_c, ignore_max=False):
    """
        Joins dataframes with df_range based on secondary key_c and value range
        df is the original dataframe
        df_range is the lookup dataframe, with unique key df_range_key
        Both df and df_range have rows split in subsets defined by secondary key key_c
        Joins column v_c in df into range values in df_range v1_c, v2_c
        If ignore_max=True v2_c is ignored 
        Returns joined dataframe, and three dataframe with missed keys, missed by range min and max
    """
    
    # Reset Index of df if there is an index and index df_range by key_c if not yet done
    df['__idx__'] = df.index
    df.set_index(['__idx__'])
    if df_range.index.names != [key_c]:
        if df_range.index.names != [None]:
            df_range = df_range.reset_index()
        df_range = df_range.set_index([key_c])
        
    # Create two columns with the min and max values for every key_c subset in df_range
    df_limits = df_range[[v1_c, v2_c]].groupby(level=[key_c])\
        .agg({v1_c: 'min', v2_c: 'max'}) \
        .rename(columns={v1_c: '__min', v2_c: '__max'})
    df_range = df_range.join(df_limits)
    df2 = df.join(df_range[[df_range_key, v1_c, v2_c, '__min', '__max']], on=[key_c])
    df_miss = df[~df[key_c].isin(df_range.index)].drop(columns=['__idx__'])
    df_miss_min = df2[df2[v_c]<df2['__min']]
    df_miss_min = df_miss_min.drop_duplicates(subset=['__idx__']).drop(columns=['__idx__'])
    df_miss_max = df2[df2[v_c]>df2['__max']]
    df_miss_max = df_miss_max.drop_duplicates(subset=['__idx__']).drop(columns=['__idx__'])
    df = df2[(df2[v_c]>=df2[v1_c])&(df2[v_c]<=df2[v2_c])]
    if ignore_max: df = pd.concat([df, df_miss_max])
    df = df.drop(columns=['__idx__']).copy()
    df['key'] = df['key'].astype(int)
    return df, df_miss, df_miss_min, df_miss_max

# ####################################################################################################
# excel
# ####################################################################################################

def format_sheet(ws, frmts, xls_format=None):
    ws.set_zoom(120)
    ws.freeze_panes(1,2)
    ws.set_column('A:Z', 18)
    if xls_format:
        for e in xls_format:
            frmt = frmts[e['frmt']] if 'frmt' in e else None
            ws.set_column(f"{e['col']}:{e['col']}",e['len'], frmt)
    else:
        ws.set_column('A:Z', 20)
        
def xls_create(p, dfs, sheet_names, xls_formats):
    writer = pd.ExcelWriter(p, engine='xlsxwriter', date_format='yyyy-mm-yy')
    workbook = writer.book
    frmts = {
        'C': workbook.add_format({'align': 'right', 'num_format': '€#,##0.00'}),
        'F': workbook.add_format({'align': 'right', 'num_format': '#,##0.0'})
    }
    workbook.set_size(1200, 800)
    for df,sheet_name,xls_format in zip(dfs,sheet_names,xls_formats):
        df.to_excel(writer, index=False, sheet_name=sheet_name)
        worksheet = writer.sheets[sheet_name]
        format_sheet(worksheet, frmts, xls_format=xls_format)
    writer.close()

def df_to_excel(df, p, xls_format=None):
    df = df[:300000]
    for c in ['Index','index']:
        if c in df.columns:
            df = df.drop(columns=[c])
    xls_create(p, [df], ['Prices'], [xls_format])


# ####################################################################################################
# TO BE REVIEWD
# ####################################################################################################

def df_parse_date_from_int(df, c):
    df.loc[df[c].isna(),c] = 0
    df.loc[:,c] = df.loc[:,c].astype(int)
    df.loc[:,c] = df[c].apply(lambda n: f'{int(n/10000)}-{int(n%10000/100)}-{int(n%100)}')
    df.loc[:,c] = pd.to_datetime(df[c], format='%Y-%m-%d', errors='coerce')

def df_lookup(df,c,d,na_val='Desconhecido'):
    """Changes in-place columns c of dataframe df with lookup values defined on dictionary d and na_val if not found"""
    df.loc[df[c].isna(),c] = na_val
    df.loc[:,c] = df[c].apply(lambda v: d[v] if v in d else na_val)

def df_to_int(df, c, na_val=-1):
    """Changes float col to int replacing NA with na_val"""
    df.loc[:,c] = df[c].fillna(na_val)
    df.loc[:,c] = df.loc[:,c].astype(int)

def df_cats_to_str(df):
    """Converts all cat cols to str"""
    for c in df.select_dtypes(include=['category']).dtypes.index:
        df[c] = df[c].astype(str)

# Change to operate inplace
def df_reduce_cats(df, c, freq=10):
    """Keeps only the values of cat col that have more than freq occurrencies"""
    dfc = pd.DataFrame(df[c].value_counts()).reset_index()
    dfc.columns = ['__key', '__val']
    dfc = dfc[dfc['__val']>freq]
    dfc = dfc.set_index(['__key'])
    dfc['__val'] = dfc.index
    df = df.join(dfc, how='left', on=c)
    df = df.drop([c], axis=1)
    df = df.rename(columns={'__val': c})
    return df

def df_cut_numeric(df, c, bins, na_val=-1):
    bins = [na_val-1] + bins
    labels = ['NA'] + [str(e) for e in bins[1:-1]]
    df_to_int(df, c)
    df.loc[:,c+'_cat'] = pd.cut(df.loc[:,c], bins, labels=labels)
    df.loc[:,c+'_cat'] = df.loc[:,c+'_cat'].astype(str)

DATE_PARTS = ['Year', 'Month', 'Week', 'Day', 'Dayofweek', 'Dayofyear', 'Is_month_end', 'Is_month_start', 
        'Is_quarter_end', 'Is_quarter_start', 'Is_year_end', 'Is_year_start']

def datepart_cols(fldname):
    fldname2_pre = re.sub('[Dd]ate$', '', fldname)
    return [f'{fldname2_pre}_{e}' for e in DATE_PARTS] + [f'{fldname2_pre}_Elapsed']

def df_add_datepart(df, fldname, drop=True):
    fld = df[fldname]
    fld_dtype = fld.dtype
    if isinstance(fld_dtype, pd.core.dtypes.dtypes.DatetimeTZDtype):
        fld_dtype = np.datetime64
    if not np.issubdtype(fld_dtype, np.datetime64):
        df[fldname] = fld = pd.to_datetime(fld, infer_datetime_format=True)
    fldname2_pre = re.sub('[Dd]ate$', '', fldname)
    for e in DATE_PARTS:
        df[f'{fldname2_pre}_{e}'] = getattr(fld.dt,e.lower())
    df[f'{fldname2_pre}_Elapsed'] = fld.astype(np.int64) // 10**9
    if drop: df.drop(fldname, axis=1, inplace=True)
    return [fldname2_pre+'_'+e for e in DATE_PARTS]

def df_meta(df, order=False):
    data = []
    for c in df.columns:
        unq = len(df[c].cat.categories) if df[c].dtype.name == 'category' else len(df[c].unique())
        examples = '; '.join([str(e) for e in df[c].value_counts().index[:5]]) 
        data.append((c, df[c].dtypes, unq, unq/len(df), df[c].isna().sum(), df[c].isna().sum()/len(df), examples))
    df1 = pd.DataFrame(data, columns=['attr', 'type', 'uniques', 'uniques_p', 'nulls', 'nulls_p', 'examples'])
    if order: df1.sort_values(['uniques'], ascending=False, inplace=True)
    return df1

def df_meta_excel(df, fname):
    writer = pd.ExcelWriter(f'{fname}.xlsx', engine='xlsxwriter')
    df1 = df_meta(df, order=False)
    df1.to_excel(writer, 'Original order')
    df1 = df_meta(df, order=True)
    df1.to_excel(writer, 'Uniques order')
    writer.save()

def df_stats(df, c, max_rows=10):
    n = len(df)
    nna = len(df[df[c].isna()])
    nu = len(df[c].unique())
    print(f'\n\033[1mCol: {c}\033[0m')
    print(f'Total: {n:,}    NA: {nna:,} ({100*nna/n:.2f}%)'.replace(',', '.'))
    print(f'Uniques: {nu:,} ({100%nu/n:.2f}%)'.replace(',', '.'))
    s = df[c].value_counts()
    data = []
    n1, p1, p2 = 0, 0.0, 0.0
    for idx, t in s.head(max_rows).items():
        data.append([idx, t, 100*t/n, 100*t/(n-nna)])
        n1 += t
        p1 += t/n
        p2 += t/(n-nna)
    if (n-nna)-n1 > 0:
        data.append(['Etc', (n-nna)-n1, 100*(1-p1), 100*(1-p2)])
    cols = ['Value', '#', '%', '%2']
    df1 = pd.DataFrame(data, columns=cols)
    return df1

def format_chart(ax, title, ylabel, wtotal, max_x, w):
    xlabel = ''
    sns.set(font_scale=1)
    title = '' if title is None else title    
    xlabel = '' if xlabel is None else xlabel
    ylabel = '' if ylabel is None else ylabel
    ax.set_title(title, fontsize=12)
    ax.set_xlabel(xlabel, fontsize=12, rotation='vertical')
    ax.set_ylabel(ylabel, fontsize=12, rotation='vertical')
    ax.get_xaxis().set_major_formatter(matplotlib.ticker.FuncFormatter(lambda x, p: f'{int(x):,}'.replace(',', '.')))
    for p in ax.patches:
        p.set_height(1)
        width = p.get_width()
        ax.text(width+max_x*0.02*(15/w), p.get_y()+p.get_height()*0.7, '{:1.2f}'.format(width/wtotal), ha='center')
    plt.subplots_adjust(hspace = 1.0)

def df_hist(df, col, n=None):
    """Calc hist with the n largest bins, or up to 99% of values if n is None"""
    x = df[col].value_counts().values
    y = list(df[col].value_counts().index.values)
    if n is None:
        total, t = sum(x), 0
        for k,e in enumerate(x):
            t += e
            if t>=total*0.99:
                break
        k += 1
        n = k if len(df[col].value_counts())>k else len(df[col].value_counts())
    if n<len(df[col].value_counts()):
        r = sum(x[n:])
        x, y = x[:n], y[:n]
        x, y = np.append(x, [r]), np.append(y, ['Etc'])
        n += 1
    return x, y, n

def df_plot(df, cols, n=None, grid_cols=2, w=15, as_cat=False, stats=False):
    if not isinstance(cols, list):
        cols, grid_cols = [cols], 1
        w = 8
    if stats:
        for col in cols:
            df1 = df_stats(df, col)
            display(df1)
    grid_rows = len(cols)//grid_cols
    nrows = []
    is_cat = df[col].dtype in ['object'] or as_cat or df[col].nunique()<10
    for i,col in enumerate(cols):
        if is_cat: _, _, n1 = df_hist(df, col, n)
        else: n1 = 10
        nrows.append(n1)
    n1 = 0
    ncols = len(cols)//2
    if ncols==0: ncols+=1
    for i in range(ncols): n1 += max(nrows[i], nrows[i]+1)
    h = (n1-1)*0.3 + 1.
    fig = plt.figure(figsize=(w, h))
    for i,col in enumerate(cols):
        ax = plt.subplot(grid_rows,grid_cols,i+1)
        if is_cat:
            x, y, _ = df_hist(df, col, n)
            ax = sns.barplot(y=y, x=x, orient='h')
            format_chart(ax, col, '#', float(len(df)), max(x)*grid_cols, w)
        else:
            bins, kde = 'auto', True
            if pd.api.types.is_integer_dtype(df[col]):
                nu = df[col].nunique()
                bins = nu if nu<20 else 20
                kde = False
            ax = sns.histplot(df[col], bins=bins, kde=kde)
        plt.gca().set_facecolor((1.0, 1.0, 1.0))
    plt.tight_layout()
    plt.show()
    plt.pause(.01)
    # plt.close('all')

# ####################################################################################################
# df_to_excel OLD
# ####################################################################################################

# def get_name(d, c): return oget(d,['cols',c,'name']) or c.capitalize()
# def get_type(d, c): return oget(d,['cols',c,'type']) or 'S'
# def get_col_w(d, c): return oget(d,['cols',c,'sz']) or 20
# def get_align(d, col_type): return 'right' if col_type in ['N', 'C', 'F'] else 'left'
# def get_num_format(d, col_type):
#     if col_type=='C': return '€# ##0.00'
#     elif col_type=='F': return '#,##0.0'
#     else: return '# ##0'
#     return '# ##0'

# def xls_format_sheet(wb, ws, d, fixed_panes, cols):
#     ALPHA = [chr(i) for i in range(65,91)]
#     ws.set_zoom(120)
#     if fixed_panes:
#         ws.freeze_panes(1,2)
#     ws.set_column('A:Z', 12)   
#     for i,c in enumerate(cols):
#         name = get_name(d, c)
#         col_type = get_type(d, c)
#         num_format = get_num_format(d, col_type)
#         align = get_align(d, col_type)
#         options = {'num_format': num_format, 'align': align}
#         frmt = wb.add_format(options)
#         ws.set_column(f'{ALPHA[i]}:{ALPHA[i]}', get_col_w(d,c), frmt)
#         options['bold'] = True
#         options['text_wrap'] = True
#         options['border'] = 0
#         options['valign'] = 'top'
#         frmt = wb.add_format(options)    
#         ws.write(0, i, name, frmt)
        
# def conditional_format_example(wb, ws, df):
#     frmt1 = wb.add_format({'bg_color': '#FFC7CE', 'font_color': '#9C0006'})
#     frmt2 = wb.add_format({'bg_color': '#C6EFCE', 'font_color': '#006100'})
#     ws.conditional_format('B2:B5', {'type': 'cell', 'criteria': '>=', 'value': 1000, 'format': frmt1})
#     ws.conditional_format('B2:B5', {'type': 'cell', 'criteria': '<=', 'value': 1000, 'format': frmt2})   

# def df_to_excel(p, dfs, sheet_names=None, d=None, fixed_panes=None, cb=None):
#     """
#         Converts dfs to an Excel file p with optional sheet_names and optional fixed panes.
#         d is the dictionary defining all names, formats and sizes of the dataframe columns.
#         Optional cb is used for conditional format (example above).
#         Doc: https://xlsxwriter.readthedocs.io
#     """
#     writer = pd.ExcelWriter(p, engine='xlsxwriter', date_format='yyyy-mm-yy')
#     wb = writer.book
#     wb.set_size(1500, 1000)
#     for i,df in enumerate(dfs):
#         for c in ['Index','index']:
#             if c in df.index:
#                 df = df.drop(columns=[c])
#         sheet_name = sheet_names[i] if sheet_names is not None and len(sheet_names)>i else f'Sheet_{i+1}'
#         df.to_excel(writer, index=False, sheet_name=sheet_name)
#         ws = writer.sheets[sheet_name]
#         xls_format_sheet(wb, ws, d, fixed_panes, df.columns)
#         if cb is not None:
#             cb(wb, ws, df)
#     writer.close()
#     return None